import 'package:curl_logger_dio_interceptor/curl_logger_dio_interceptor.dart';
import 'package:d_dio/src/type/type.dart';
import 'package:d_dio/src/utils/log.dart';
import 'package:d_dio/src/web/dio_client.dart';
import 'package:dio/dio.dart';
import 'package:dio/native_imp.dart';
import 'package:dio_log/dio_log.dart';
import 'package:flutter/cupertino.dart';

import 'handler/data_error_handler.dart';
import 'interceptor/cache_interceptor.dart';
import 'interceptor/cookie_interceptor.dart';
import 'interceptor/error_interceptor.dart';
import 'interceptor/formdata_interceptor.dart';
import 'interceptor/log_interceptor.dart';
import 'interceptor/request_interceptor.dart';
import 'interceptor/sign_interceptor.dart';
import 'options/request_options_extra.dart';
import 'transformer/json_transformer.dart';

final Http http = Http();

///  Http中的泛型T代表BaseResp的data类型、不传泛型则返回原数据
///  Stream或Future返回的数据均为success为true的BaseResp中的data
///  错误处理需在Stream的onError或Future的catchError
class Http extends DioForNative {
  static Http instance = Http._();

  factory Http() => instance;

  Http._() {
    transformer = JsonTransformer();
  }

  bool _isLog = false;

  Http isLog(bool log) {
    _isLog = isDebug(log);
    return this;
  }

  Http proxy(String? proxy) {
    setDioClient(this, proxy: proxy);
    return this;
  }

  String _wsUrl = '';

  Http baseUrl(String url, String urlForWs) {
    options.baseUrl = url;
    _wsUrl = urlForWs;
    return this;
  }

  Http timeout(int time) {
    options
      ..connectTimeout = time
      ..receiveTimeout = time;
    return this;
  }

  String _secret = '';
  String _newSecret = '';

  Http signSecret(String signSecret, String newSignSecret) {
    _secret = signSecret;
    _newSecret = newSignSecret;
    return this;
  }

  CacheReader? _cacheReader;

  Http cacheReader(CacheReader reader) {
    _cacheReader = reader;
    return this;
  }

  CacheSaver? _cacheSaver;

  Http cacheSaver(CacheSaver saver) {
    _cacheSaver = saver;
    return this;
  }

  JsonConvert? _jsonConverter;

  Http jsonConverter(JsonConvert converter) {
    _jsonConverter = converter;
    return this;
  }

  ErrorHandler? _errorHandler;

  Http errorHandler(ErrorHandler handler) {
    _errorHandler = handler;
    return this;
  }

  void showLogBtn(BuildContext context){
    if (_isLog) {
      WidgetsBinding.instance.addPostFrameCallback((_) {
        showDebugBtn(context);
      });
    }
  }

  /// call before need call baseUrl、signSecret、errorHandler、cacheSaver
  Http addInterceptors([List<Interceptor> values = const []]) {
    interceptors.addAll([
      CookieInterceptor(),
      RequestInterceptor(wsUrl: _wsUrl, cacheReader: _cacheReader),
      SignInterceptor(secret: _secret, newSecret: _newSecret),
      FormDataInterceptor(),
      ErrorInterceptor(errorHandler: _errorHandler),
      CacheInterceptor(cacheSaver: _cacheSaver),
      LogsInterceptor(isLog: _isLog, cacheSaver: _cacheSaver),
      if (_isLog) DioLogInterceptor(),
      if (_isLog) CurlLoggerDioInterceptor(printOnSuccess: true),
      ...values,
    ]);

    return this;
  }

  Future<T> requestWrap<T>(
    String path, {
    data,
    Map<String, dynamic>? queryParameters,
    Options? options,
    CancelToken? cancelToken,
    ProgressCallback? onSendProgress,
    ProgressCallback? onReceiveProgress,
    //extra
    bool? needToken,
    bool? wsApi,
    bool? openCache,
    bool? useFormData,
  }) {
    var optionsWrap = RequestOptionsExtra.compose(options,
        needToken: needToken,
        wsApi: wsApi,
        openCache: openCache,
        useFormData: useFormData);
    return request(path,
            data: data,
            queryParameters: queryParameters,
            cancelToken: cancelToken,
            options: optionsWrap,
            onSendProgress: onSendProgress,
            onReceiveProgress: onReceiveProgress)
        .then((v) => onResponse(v, _jsonConverter!), onError: onError);
  }

  ///同步请求 需添加cancelToken dispose时取消
  Future<T> getSync<T>(String path,
          {Map<String, dynamic>? params,
          Options? options,
          bool? needToken,
          bool? wsApi,
          bool? openCache,
          CancelToken? cancelToken,
          ProgressCallback? onReceiveProgress}) =>
      requestWrap(path,
          options: DioMixin.checkOptions('GET', options),
          queryParameters: params,
          needToken: needToken,
          wsApi: wsApi,
          openCache: openCache,
          cancelToken: cancelToken,
          onReceiveProgress: onReceiveProgress);

  Future<T> postSync<T>(String path,
          {data,
          Options? options,
          bool? needToken,
          bool? wsApi,
          bool? openCache,
          bool? useFormData,
          Map<String, dynamic>? queryParameters,
          CancelToken? cancelToken,
          ProgressCallback? onSendProgress,
          ProgressCallback? onReceiveProgress}) =>
      requestWrap(path,
          options: DioMixin.checkOptions('POST', options),
          data: data,
          queryParameters: queryParameters,
          needToken: needToken,
          wsApi: wsApi,
          useFormData: useFormData,
          openCache: openCache,
          cancelToken: cancelToken,
          onSendProgress: onSendProgress,
          onReceiveProgress: onReceiveProgress);

  ///异步请求 dispose时取消订阅
  Stream<T> getRx<T>(String path,
          {Map<String, dynamic>? params,
          Options? options,
          bool? needToken,
          bool? wsApi,
          bool? openCache,
          CancelToken? cancelToken,
          ProgressCallback? onReceiveProgress}) =>
      requestWrap<T>(path,
              options: DioMixin.checkOptions('GET', options),
              queryParameters: params,
              needToken: needToken,
              wsApi: wsApi,
              openCache: openCache,
              onReceiveProgress: onReceiveProgress)
          .asStream();

  Stream<T> postRx<T>(String path,
          {data,
          Options? options,
          bool? needToken,
          bool? wsApi,
          bool? openCache,
          bool? useFormData,
          Map<String, dynamic>? queryParameters,
          CancelToken? cancelToken,
          ProgressCallback? onSendProgress,
          ProgressCallback? onReceiveProgress}) =>
      requestWrap<T>(path,
              options: DioMixin.checkOptions('POST', options),
              data: data,
              queryParameters: queryParameters,
              needToken: needToken,
              wsApi: wsApi,
              useFormData: useFormData,
              openCache: openCache,
              onSendProgress: onSendProgress,
              onReceiveProgress: onReceiveProgress)
          .asStream();
}
